#include <cassert>
#include <iostream>
#include <string>
#include <utility> // 'std::swap'

#include "Human_being.h"

const std::string Human_being::_Default_nickname = "?";

Human_being Human_being::Default_Human_being() {
	return Human_being(_Default_nickname); // Created object *IS COPIED* to caller context
}

Human_being::Human_being(const std::string& nickname) : nickname(nickname) {
	_resource = new Resource();
}

Human_being::Human_being(const Human_being& hb) : nickname(hb.nickname) {
	_resource = new Resource(); // To be more realistic: copy data from 'hb' as well...
}

Human_being::Human_being(Human_being&& hb) noexcept : nickname(hb.nickname) {
	std::cout << "Move constructor from: " << hb.nickname << std::endl;
	// Move semantics:
	std::swap(_resource, hb._resource); // 'delete _resource;' on 'hb' is later done when "rvalue" is destroyed...
}

Human_being::~Human_being() {
	delete _resource;
}

Human_being& Human_being::operator=(Human_being const& hb) {
	const_cast<std::string&> (this->nickname) = hb.nickname;
	if (this != &hb) { // Weird case: 'Human_being p; p = p;'
		delete _resource;
		_resource = new Resource(); // To be more realistic: copy data from 'hb' as well...
	}
	return *this;
}

Human_being& Human_being::operator=(Human_being&& hb) noexcept {
	std::cout << "Move assignment from: " << hb.nickname << std::endl;
	const_cast<std::string&> (this->nickname) = hb.nickname;
	// Move semantics:
	std::swap(_resource, hb._resource); // 'delete _resource;' on 'hb' is later done when "rvalue" is destroyed...
	return *this;
}

void Human_being::becomes(Human_being&& hb) { // 'hb' is "lvalue"!
	// Rule: "Things that are declared as rvalue reference can be lvalues or rvalues.
	// The distinguishing criterion is: if it has a name then it is an lvalue else it is an rvalue."
	*this = hb; // 'Human_being& Human_being::operator=(Human_being const& hb)' is called...
}

Resource* const Human_being::get_resource() const {
	return _resource;
}

int main(int argc, char** argv) {
	// Compilation error:
	// Human_being& who = Human_being("who"); // A "lvalue" cannot bind a "temporary" object, i.e., a "rvalue"...
	// Copy elision through optimization, i.e., the move constructor *IS NOT* called:
	Human_being&& who = Human_being("who"); // While a "rvalue" can!

	// The 'Human_being' object returned by 'Human_being::Default_Human_being' is *MOVED* from callee context to caller context
	// *WITHOUT* superfluous copy...
	// This often was a compilation optimization that became a developer-controlled mechanism from C++11:
	Human_being&& default_Human_being = Human_being::Default_Human_being(); // Debugger shows that copy *ISN'T* called...

	Human_being someone;
	// 'static Human_being GetInstance(const std::String& s) { Human_being hb(s); return hb; } // Deletion of 'hb'...
	// "Move constructor" (return) and next "move assignment":
	someone = Human_being::GetInstance("someone else...");

	Human_being EmmanuelMacron("EmmanuelMacron");
	Human_being president;
	Resource* resource = president.get_resource();
	president = std::move(EmmanuelMacron); // Move assignment ('_resource' fields are exchanged...)
	assert(EmmanuelMacron.get_resource() == resource);

	Human_being anybody;
	anybody.becomes(Human_being::GetInstance("anyone"));

	// 'GetClone(Human_being& hb)' does not work:
	// Human_being attendee = Human_being::GetClone(Human_being::GetInstance("attendee")); // Compilation error: "lvalue" required...
	// 'GetClone(const Human_being& hb)' makes it work;
	Human_being attendee = Human_being::GetClone(Human_being::GetInstance("attendee")); // "rvalue" "disappears"...
	Human_being attendee_ = Human_being::GetClone(Human_being::GetInstance("attendee_")); // "rvalue" passed, OK...

	Human_being FranckBarbier("FranckBarbier");
	Human_being& Bab = FranckBarbier; // Copy constructor ("lvalue"): '_resource' is shared...
	assert(FranckBarbier.get_resource() != nullptr && Bab.get_resource() != nullptr && FranckBarbier.get_resource() == Bab.get_resource());

	Human_being trainer = ::GetClone/*<Human_being&>*/(FranckBarbier); // "lvalue"
	assert(FranckBarbier.get_resource() != nullptr && trainer.get_resource() != nullptr && FranckBarbier.get_resource() != trainer.get_resource());

	return 0;
}

